/**************************************
* macfilter.c
*
* keep mac addresses from being sent
* out the wrong interface.  Fixes
* problems with some switches.
*************************************/

#include "macfilter.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//#define DEBUG
//#define DEBUGPATH

/*------------------globals------------*/
FilterHolder	Macs[MAX_MACS];
FilterHolder*	MacHead;

typedef struct mac_ether{
	unsigned long long	src:24;
	unsigned long long  dst:24;
	unsigned short		protocol;
} MACEther;

void mac_init(int argc, char** argv){
#ifdef DEBUGPATH
	printf("In mac_init\n");
#endif
	bzero(Macs, sizeof(FilterHolder)*MAX_MACS);
} 


void PrintMac(unsigned char* m){
	int i;

#ifdef DEBUGPATH
	printf("In PrintMac\n");
#endif
	
	for (i=0;i<6;i++){
		printf("%02x:",m[i]);
	}
	printf("\n");
}

/*TODO: Splay to the front*/
int ProcessMac(PacketToProcess* ptp){
	FilterHolder*		this;
	FilterHolder*		last;
	
#ifdef DEBUG
	printf("------------------------------\n");
#endif	
	
	if (!MacHead){
#ifdef DEBUG
		printf("Dropping Packet: First Packet\n");
#endif	
		MacHead = calloc(sizeof(FilterHolder), sizeof(char));
		MacHead->last_time=ptp->ts.tv_sec;
		memcpy(MacHead->mac, (ptp->raw_packet)+6, 6);
		MacHead->interface = ptp->from_interface;
		return TRUE;
	}
	
	this = MacHead;
	while (this){
		if (memcmp( (ptp->raw_packet)+6, this->mac, 6) ==0){
#ifdef DEBUG
			printf("Found Mac: ");
			PrintMac(this->mac);
#endif		

			/*move this one to the front so we can find it faster*/
			if (this!=MacHead){
				if (this->prev) this->prev->next = this->next;
				if (this->next) this->next->prev = this->prev;
				this->next = MacHead;
				MacHead->prev = this;
				MacHead = this;
			}

			if (this->interface != ptp->from_interface){
#ifdef DEBUG
				printf("Dropping: interface mismatch ");
				PrintMac(this->mac);
#endif			
				return TRUE;
			}
			
			this->last_time = ptp->ts.tv_sec;
			if (this->count < 50) this->count++;
#ifdef DEBUG			
			if (this->count == 3){
				printf("Finalized ");
				PrintMac(this->mac);
			}
#endif							
			return (this->count<3);				

		}
		if (ptp->ts.tv_sec - this->last_time > MAC_TIMEOUT){
			if (this->prev){
				this->prev->next = this->next;
			}else{
				MacHead = this->next;
				this->next->prev = NULL;
			}
			if (this->next){	
				this->next->prev = this->prev;
			}else{
				this->prev->next = NULL;
			}
			free(this);	
		}	
			
		last = this;
		this = this->next;
	}
	
	last->next = calloc(sizeof(FilterHolder), sizeof(char));
	last->next->prev = last;
	last = last->next;
	last->count = 1;
	memcpy(last->mac, (ptp->raw_packet) + 6 , 6);
	last->last_time = ptp->ts.tv_sec;
	last->interface = ptp->from_interface;
	
#ifdef DEBUG
	printf("Adding on interface %i ",ptp->from_interface);
	PrintMac(last->mac);
#endif	
	
	return 1;
}

int ProcessMacOld(PacketToProcess* ptp){
	int 			i;
	int 			next_free;

#ifdef DEBUGPATH
	printf("In ProcessMac\n");
#endif

	/*search for the mac*/
	next_free=-1;
	for (i=0;i<MAX_MACS;i++){
		if (Macs[i].count==0){
			if (next_free==-1) next_free=i;
		}else{
			if (memcmp((ptp->raw_packet)+6 , Macs[i].mac , 6)==0){
				if (Macs[i].interface!=ptp->from_interface){
#ifdef DEBUG				
					printf("DROPPING: interface mismatch ");
					PrintMac(Macs[i].mac);
#endif					
					return TRUE;
				}
				Macs[i].last_time=ptp->ts.tv_sec;
				if (Macs[i].count<50) Macs[i].count++;
				if (Macs[i].count==3){
#ifdef DEBUG				
					printf("(%i) Finalized ",i);
					PrintMac(Macs[i].mac);
#endif					
				}
				return (Macs[i].count<3);
			}
		}
		if (ptp->ts.tv_sec - Macs[i].last_time > MAC_TIMEOUT)
			Macs[i].count=0;
	}

	if (next_free==-1) return FALSE;	

	Macs[next_free].count=1;
	memcpy(Macs[next_free].mac, (ptp->raw_packet)+6 , 6);
	Macs[next_free].last_time=ptp->ts.tv_sec;
	Macs[next_free].interface=ptp->from_interface;

#ifdef DEBUG
	printf("Adding (%i)On interface %i  ",next_free,ptp->from_interface);
	PrintMac(Macs[next_free].mac);
#endif	

	return TRUE;
}
