/*********************************************
* Linux specific code for packet handling
**********************************************/
#include "packet.h"
#include "hogwash.h"
/*only compile under linux*/
#ifdef __LINUX__

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <sys/time.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <sys/poll.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*-------------Global Variables------------*/
int	fd1,fd2;
int 	fds1, fds2;
int	MTU1, MTU2;

/* SPAN */
int	fd3;
int	fds3;
int	MTU3;


#ifdef ERASEME
/*********************************************
* Get the MTU of interface named "name"
*********************************************/
int GetIfrMTU(char *name) {
    int fd;
    struct ifreq ifr;
    int retval;

    retval = -1;
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if ( fd < 0) {
        printf("Couldn't create socket for MTU\n");
        return -1;
    }

    strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name));
    if (ioctl(fd, SIOCGIFMTU, &ifr) == 0)
        retval = ifr.ifr_metric;
    else
        printf("ioctl(SIOCGIFMTU)");    
    close(fd);
	return retval;
}

#endif

/*********************************************
* Get the id of interface named "name"
*********************************************/
int get_device_id(int fd, char* name){
	struct ifreq	ifr;

	bzero(&ifr, sizeof(struct ifreq));
	strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name));	
	if (ioctl(fd, SIOGIFINDEX, &ifr) == -1) return -1;
	return ifr.ifr_ifindex;	
}


/********************************************
* Get the interface ready for reading
********************************************/
int open_interface(char* name){
	int 				fd;
	struct sockaddr_ll	sll;
	int					ssize;
	int					errnum;
	struct packet_mreq	mr;

#ifdef DEBUG
	printf("Opening interface %s\n", name);
#endif

	fd=socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	bzero(&sll,sizeof(struct sockaddr_ll));
	sll.sll_family=AF_PACKET;
	sll.sll_ifindex=get_device_id(fd, name);
	sll.sll_protocol=htons(ETH_P_ALL);
	ssize=sizeof(struct sockaddr_ll);

	errnum=bind(fd, (struct sockaddr*)&sll, sizeof(sll));
	if (errnum==-1){
		printf("Error Binding socket\n");
	}
	
	/*set promisc mode*/
	memset(&mr, 0, sizeof(mr));
	mr.mr_ifindex=get_device_id(fd, name);
	mr.mr_type=PACKET_MR_PROMISC;
	if (setsockopt(fd, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mr, sizeof(mr))==-1){
		printf("Failed to set promisc mode\n");
		exit(1);
	}

	return fd;	
} 

/*********************************************
* Open up the interfaces
*********************************************/
int open_interfaces(char* internalif, char* externalif, char* spanif, int use_span){

#ifdef DEBUG
	printf("Opening Interfaces for linux\n");
#endif
	fds1=open_interface(internalif);
	fds2=open_interface(externalif);
	if (use_span == TRUE) {
 	  fds3=open_interface(spanif);
	}
	MTU1=GetIfrMTU(internalif);
	MTU2=GetIfrMTU(externalif);
	if (use_span == TRUE) {
	  MTU3=GetIfrMTU(spanif);
	}

	printf("At the end of open_interfaces %i and %i\n",fds1, fds2);
	if (use_span == TRUE) {
          printf("At the end of open_interfaces %i",fds3);
	}
	fd1=fds1;
	fd2=fds2;
	if (use_span == TRUE) {
	  fd3=fds3;
	  return (fds1>0 && fds2>0 && fds3>0);
	} else {
	  return (fds1>0 && fds2>0);
	}
}

/*********************************************
* Open up the interfaces
*********************************************/
void main_loop(int use_span){
	int 			bufsize1, bufsize2, cc;
	u_char 			*buf1, *buf2;
	int 			i;
	PacketToProcess		ptp;
	
	struct timeval	tv;	
	struct pollfd	pds[2];
	int				numfds;
	
#ifdef DEBUG
	printf("Entering Main loop for linux\n");
#endif	


	printf("In main_loop %i,%i and %i,%i\n", fds1,fd1, fds2,fd2);	
	/*initialize the set*/
	pds[0].fd=fds1;
	pds[0].events=POLLIN;
	pds[0].revents=0;
	pds[1].fd=fds2;
	pds[1].events=POLLIN;
	pds[1].revents=0;

	/*Get the sizes of the buffers*/
	bufsize1=MTU1;
	if ((bufsize1 <= 0) || (bufsize1 >= 65536)) bufsize1=65536;
#ifdef DEBUG	
	printf("Buffer size1 is %i\n",bufsize1);
#endif	
	buf1 = (u_char *) malloc((unsigned) bufsize1);
	if (buf1 == 0) {
		printf("MainLoop: malloc failed\n");
		exit(1);
	}
	
	bufsize2=MTU2;
	if ((bufsize2 <= 0) || (bufsize2 >= 65536)) bufsize2=65536;
#ifdef DEBUG	
	printf("Buffer size1 is %i\n",bufsize2);
#endif	
	buf2 = (u_char *) malloc((unsigned) bufsize2);
	if (buf2 == 0) {
		printf("MainLoop: malloc failed\n");
		exit(1);
	}
	/*end getting the sizes of the buffers*/

	
	while (1){
		tv.tv_sec=1;
		tv.tv_usec=0;
		numfds=poll(pds, 2, 1);
		
		if (numfds>0){
			for (i=0;i<2;i++){
				if ( (pds[i].revents & POLLIN)== POLLIN ){
#ifdef DEBUG				
					printf("Got packet on interface %i\n", i);
#endif					
					bzero(&ptp, sizeof(PacketToProcess));
					ptp.from_interface=i;
					cc = read(pds[i].fd, (char *) ptp.raw_packet, PACKET_HOLDER_SIZE);
					ptp.packet_len=cc;
					ioctl(pds[i].fd, SIOCGSTAMP, &ptp.ts);
#ifdef DEBUG					
					printf("Read %i bytes from the socket\n",cc);
					printf("tv.tv_sec %li\n",ptp.ts.tv_sec);
					printf("packet_len is %u\n",ptp.packet_len);				
#endif						
					HogwashProcessPacket(&ptp, use_span);
				}
			}
		}
	}
}

/*********************************************
* Open up the interfaces
*********************************************/
void restart(){
}

/*********************************************
* Open up the interfaces
*********************************************/
void close_interfaces(int use_span){
	close(fds1);
	close(fds2);
	if (use_span == TRUE) {
          close(fds3);
	}
}

/*********************************************
* Write a packet to the interface
*********************************************/
int	 write_packet(int interface, char* packet, int packet_len, int use_span){
	int cc;
	int	fd;
	
	if (interface==INTERNAL)
		fd=fds1;
	else
		fd=fds2;

	if (use_span == TRUE) {
	  cc=write(fds3, packet, packet_len);
          if (cc!=packet_len){
		printf("write_packet: span packet copy size did not match\n");
	  }
	}

	cc=write(fd, packet, packet_len);
	if (cc!=packet_len){
#ifdef DEBUG	
		printf("write_packet: sizes did not match\n");
#endif			
		return FALSE;
	}
	return TRUE;
}

#endif
