#include "pass_rule.h"
/**********************************
* pass_rule.c
*
* if pass_rule return TRUE, the 
* packet is not filtered.  If 
* it returns FALSE the packet is
* filtered
**********************************/
#include "../snort-1.8.6/snort.h"
#include "../snort-1.8.6/decode.h"
#include "../snort-1.8.6/rules.h"
#include "../snort-1.8.6/parser.h"
#include "../snort-1.8.6/checksum.h"
#include "packet.h"
#include <dirent.h>
#include <sys/types.h>

extern ListHead* head_tmp;
extern PV pv;
extern RuleListNode* RuleLists;

#ifdef __USE_PTHREADS__
pthread_mutex_t snort_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif


/*structs used to build return packets*/
typedef struct h_ethernet
{
    u_int8_t ether_dst[6];
    u_int8_t ether_src[6];
    u_int16_t ether_type;

}HEthernet;

typedef struct h_ip
{
    u_int8_t ip_hlen:4, ip_ver:4;
    u_int8_t ip_tos;        /* type of service */
    u_int16_t ip_len;       /* datagram length */
    u_int16_t ip_id;        /* identification  */
    u_int16_t ip_off;       /* fragment offset */
    u_int8_t ip_ttl;        /* time to live field */
    u_int8_t ip_proto;      /* datagram protocol */
    u_int16_t ip_csum;      /* checksum */
    struct in_addr ip_src;  /* source IP */
    struct in_addr ip_dst;  /* dest IP */
}HIP;

typedef struct h_tcp
{
    u_int16_t th_sport;     /* source port */
    u_int16_t th_dport;     /* destination port */
    u_int32_t th_seq;       /* sequence number */
    u_int32_t th_ack;       /* acknowledgement number */
    u_int8_t th_x2:4,th_off:4;
    u_int8_t th_flags;
    u_int16_t th_win;       /* window */
    u_int16_t th_sum;       /* checksum */
    u_int16_t th_urp;       /* urgent pointer */

}HTCP;

#define	H_TCP	6


/********************************************************
* Create a reset packet to tear down the connection
*
* returns: number of bytes in the new packet
           pkt contains the new bytes
********************************************************/
int CreateResetPacket(PacketToProcess* ptp, Packet* p, int Span){
    struct pseudoheader       /* pseudo header for TCP checksum calculations */
    {
        u_int32_t sip, dip;   /* IP addr */
        u_int8_t  zero;       /* checksum placeholder */
        u_int8_t  protocol;   /* protocol number */
        u_int16_t tcplen;     /* tcp packet length */
    };

	unsigned char 	Transport;
	char			NewBuff[PACKET_HOLDER_SIZE];
	int				PacketLen;
	int				to_interface;
	char*			OldPacket;
	
	HEthernet	*hethn, *hetho;
	HIP			*hipn, *hipo;
	HTCP		*htcpn, *htcpo;
	
	struct pseudoheader	ph;

	bzero(NewBuff,PACKET_HOLDER_SIZE);
	if (ptp->from_interface==0) to_interface=1; else to_interface=0;
	OldPacket=ptp->raw_packet;

	/*determine transport layer*/
	Transport=*(unsigned char*)(OldPacket+23);
	if (Transport!=H_TCP) return FALSE;
	
	/*cast struct over everything*/
	hetho=(HEthernet*)OldPacket;
	hethn=(HEthernet*)NewBuff;
	hipo=(HIP*)(OldPacket+sizeof(HEthernet));
	hipn=(HIP*)(NewBuff+sizeof(HEthernet));
	htcpo=(HTCP*)(OldPacket+sizeof(HEthernet)+sizeof(HIP));
	htcpn=(HTCP*)(NewBuff+sizeof(HEthernet)+sizeof(HIP));

	/***********************************/
	/* Create packet to send back      */
	/***********************************/
	
	/*build the ethernet header*/
	memcpy(hethn->ether_dst, hetho->ether_src, 6);
	memcpy(hethn->ether_src, hetho->ether_dst, 6);
	hethn->ether_type=hetho->ether_type;
	/*done building ethernet header*/
	
	/*build the ip header*/
	hipn->ip_ver=4;
	hipn->ip_hlen=5;
	hipn->ip_tos=0;
	hipn->ip_len=htons(sizeof(HIP)+sizeof(HTCP));
	hipn->ip_id=555*rand();
	hipn->ip_off=0;
	hipn->ip_ttl=255;
	hipn->ip_proto=6;
	hipn->ip_csum=0;
	hipn->ip_src=hipo->ip_dst;
	hipn->ip_dst=hipo->ip_src;
	/*done building ip header*/
	
	/*build the transport header*/
	htcpn->th_sport=htcpo->th_dport;
	htcpn->th_dport=htcpo->th_sport;
	htcpn->th_seq=htcpo->th_ack;
	htcpn->th_ack=htonl(ntohl(htcpo->th_seq)+1);
	htcpn->th_x2=5;
	htcpn->th_off=0;
	htcpn->th_flags=4;
	htcpn->th_win=128;
	htcpn->th_sum=0;
	htcpn->th_urp=0;
	PacketLen=sizeof(HEthernet)+sizeof(HIP)+sizeof(HTCP);
	/*done building the transport header*/
	
	/*calculate the checksums*/
	hipn->ip_csum=checksum((u_short *)hipn, sizeof(HIP), (u_short *)NULL, 0);
	ph.sip = (u_int32_t)(hipn->ip_src.s_addr);
	ph.dip = (u_int32_t)(hipn->ip_dst.s_addr);
	ph.zero=0;
	ph.protocol=6;
	ph.tcplen = htons(sizeof(HTCP));
	htcpn->th_sum=checksum((u_int16_t*)&ph, 12 , (u_int16_t*) htcpn, sizeof(HTCP));
	/*done calculating the checksums*/
	
	write_packet(to_interface, NewBuff, PacketLen, Span);
	
	/***********************************/
	/* Create packet to send on        */
	/***********************************/
	bzero(NewBuff,PACKET_HOLDER_SIZE);

	/*build the ethernet header*/
	memcpy(hethn->ether_dst, hetho->ether_dst, 6);
	memcpy(hethn->ether_src, hetho->ether_src, 6);
	hethn->ether_type=hetho->ether_type;
	/*done building ethernet header*/
	
	/*build the ip header*/
	hipn->ip_ver=4;
	hipn->ip_hlen=5;
	hipn->ip_tos=0;
	hipn->ip_len=htons(sizeof(HIP)+sizeof(HTCP));
	hipn->ip_id=hipo->ip_id;
	hipn->ip_off=0;
	hipn->ip_ttl=255;
	hipn->ip_proto=6;
	hipn->ip_csum=0;
	hipn->ip_src=hipo->ip_src;
	hipn->ip_dst=hipo->ip_dst;
	/*done building ip header*/
	
	/*build the transport header*/
	htcpn->th_sport=htcpo->th_sport;
	htcpn->th_dport=htcpo->th_dport;
	htcpn->th_seq=htcpo->th_seq;
	htcpn->th_ack=htcpo->th_ack;
	htcpn->th_x2=5;
	htcpn->th_off=0;
	htcpn->th_flags=4;
	htcpn->th_win=128;
	htcpn->th_sum=0;
	htcpn->th_urp=0;
	PacketLen=sizeof(HEthernet)+sizeof(HIP)+sizeof(HTCP);
	/*done building the transport header*/
	
	/*calculate the checksums*/
	hipn->ip_csum=checksum((u_short *)hipn, sizeof(HIP), (u_short *)NULL, 0);
	ph.sip = (u_int32_t)(hipn->ip_src.s_addr);
	ph.dip = (u_int32_t)(hipn->ip_dst.s_addr);
	ph.zero=0;
	ph.protocol=6;
	ph.tcplen = htons(sizeof(HTCP));
	htcpn->th_sum=checksum((u_int16_t*)&ph, 12 , (u_int16_t*) htcpn, sizeof(HTCP));
	/*done calculating the checksums*/
	
	write_packet(ptp->from_interface, NewBuff, PacketLen, Span);
	
	return TRUE;
}

/*****************************************************
* Parses the rule and output the sid line
******************************************************/
int write_sid_line(char* rawrule, FILE* sidfp){
	char	msg[65535];
	int		sid;
	char*	startsid;
	char*	startmsg;
	char*	endmsg;
	
	if (*rawrule==0x00 || *rawrule=='#') return TRUE;
	
#ifdef DEBUG	
	printf("Rule is\n%s\n",rawrule);
#endif	
	
	/*find the SID*/
	startsid=strstr(rawrule, "sid:");
	if (!startsid){
		printf("Couldn't find sid\n");
		return FALSE;
	}	
	sid = atoi(startsid+4);
	
	/*find the msg*/
	startmsg=strstr(rawrule, "msg:");
	if (!startmsg){
		printf("Couldn't find msg\n");
		return FALSE;
	}
	startmsg=startmsg+4;
	endmsg=strstr(rawrule, ";");
	if (!endmsg){
		printf("Couldn't find end of msg\n");
		return FALSE;
	}
	bzero(msg, 65535);
	memcpy(msg, startmsg, endmsg-startmsg);
	
	/*write the sid line*/
	fprintf(sidfp, "%i || %s\n", sid, msg);

	return TRUE;	
}

/*****************************************************
* Generate and SID file based on the current rules file
******************************************************/
int generate_sid_file(char* rulepath, char* sidpath){
	FILE*	rulefp;
	FILE*	sidfp;
	char	RuleBuff[1024];
	int		count;
	int		LineCount;
	char*	CurLine;
	char*	CurPos;
	
#ifdef DEBUG
	printf("In generate_sid_file\n");
	printf("Rules File: %s\n",rulepath);
	printf("SID   File: %s\n",sidpath);
#endif
	
	rulefp=fopen(rulepath, "r");
	if (!rulefp){
		printf("Error Opening Rules File\n");
		return FALSE;
	}
	sidfp=fopen(sidpath, "w");
	if (!sidfp){
		printf("Error Opening SID File\n");
		return FALSE;
	}	
	
	bzero(RuleBuff, 1024);
	CurLine=RuleBuff;
	CurPos=RuleBuff;
	count=0;
	LineCount=1024;
	while ( (count=fread(CurPos, 1,LineCount, rulefp)) ){
		CurPos=RuleBuff;
		while (*CurPos!=0x0A && *CurPos!=0x00) CurPos++;
		*CurPos=0x0;
		CurPos++;
		write_sid_line(CurLine, sidfp);
		LineCount=CurPos-RuleBuff;
		memcpy(RuleBuff, CurPos, 1024-LineCount);
		CurPos=RuleBuff+1024-LineCount;
		bzero(CurPos, LineCount);
	}
	
	/*process the rest of the buffer*/
	CurPos=RuleBuff;
	CurLine=RuleBuff;
	LineCount=1;
	while (CurPos< (RuleBuff+1024)){
		CurLine=CurPos;
		while (*CurPos!=0x0A && *CurPos!=0x00) CurPos++;
		*CurPos=0x0;
		CurPos++;
		write_sid_line(CurLine, sidfp);
		LineCount=CurPos-CurLine;
	}	
	
	fclose(rulefp);
	fclose(sidfp);
	
	return TRUE;
}

/*****************************************************
* find the highest numbered rules file in the given dir
******************************************************/
int get_last_rules_filename(char* subdir, char* name, int namelen){
	DIR*			dir;
	struct dirent*	d;
	int				biggest;
	int				lastnum;
	
	dir=opendir(subdir);
	if (!dir){
		printf("Couldn't open subdir %s\n",subdir);
		return FALSE;
	}
	
	biggest=-1;
	while ((d=readdir(dir))){
		if (strncmp(d->d_name, "hogwash-rules.v", 7)==0){
			lastnum=atoi(d->d_name+7);
			if (lastnum>biggest){
				sprintf(name, "%s/%s", subdir, d->d_name);
				biggest=lastnum;
			}	
		}
	}

	printf("Using filename %s\n",name);	
	closedir(dir);
	return (biggest!=-1);
}

/*****************************************************
* read the command line for anything that belonds to this
* pass rule
******************************************************/
void pass_rule_cmd_line(int argc, char* argv[]){
	int 	i=0;
	int		GenerateSid;
	
#ifdef DEBUG
	printf("In pass_rule_cmd_line\n");
#endif	
	
	GenerateSid=FALSE;
	printf("SidPath is %p\n",SidPath);
	bzero(SidPath, MAX_FILENAME_SIZE);
	for (i=0; i<argc; i++){
#ifdef DEBUG
		printf("Parsing arg %s\n",argv[i]);
#endif		
		if (strncmp("-C", argv[i], 2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				printf("Syntax -C <rules file dir>\n");
				exit(1);
			}
			if (!get_last_rules_filename(argv[i+1], pv.config_file, STD_BUF-1)){
				printf("Couldn't get rules filename\n");
				exit(1);
			}
			pv.use_rules=1;
			pv.rules_order_flag=1;
			OrderRuleLists("pass alert log");
			ParseRulesFile(pv.config_file, 0);
		}else if (strncmp("-c", argv[i], 2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				printf("Syntax -c <rules file>\n");
				exit(1);
			}
			strncpy(pv.config_file, argv[i+1], STD_BUF-1);
			pv.use_rules=1;
			pv.rules_order_flag=1;
			OrderRuleLists("pass alert log");
			ParseRulesFile(pv.config_file, 0);
		}else if (strncasecmp("-l", argv[i], 2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				printf("Syntax -l <log directory>\n");
				exit(1);
			}
			strncpy(pv.log_dir, argv[i+1], STD_BUF-1);
			pv.log_flag=1;
		}else if (strncasecmp("--sid", argv[i], 5)==0){
#ifdef DEBUG
			printf("Will Generate a SID file\n");
#endif		
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				printf("Syntax --sid <new sid path>\n");
				exit(1);
			}
			snprintf(SidPath, 65535, argv[i+1]);
			GenerateSid=TRUE;
		}

	}
	
	if (!pv.use_rules){
		printf("You must specify a rules file with -c <rules file>\n");
		exit(1);
	}
	
	if (GenerateSid){
#ifdef DEBUG
		printf("Generating SID file\n");
#endif	
		generate_sid_file(pv.config_file, SidPath);
	}
}

/*********************************************************
* Add the rules we can't work without to snort
*********************************************************/
void add_mandatory_rules(){
	char		Rule[1024];
	ListHead*	listhead=NULL;

#ifdef DEBUG	
	printf("***********Adding mandatory rules****************\n");
#endif	
	sprintf(Rule,"output scrub_alert: 0\n");
   	/*ParseRule(Rule,0);*/
	
	listhead = CreateRuleType("drop", RULE_ALERT, 1, NULL);
	head_tmp = listhead;
	ParseOutputPlugin(Rule);
	head_tmp=NULL;
	
	sprintf(Rule,"output scrub_alert: 1\n");
		
	listhead = CreateRuleType("sdrop", RULE_ALERT, 1, NULL);
	head_tmp = listhead;
	ParseOutputPlugin(Rule);
	head_tmp=NULL;

	sprintf(Rule,"output scrub_alert: 2\n");
		
	listhead = CreateRuleType("ignore", RULE_ALERT, 1, NULL);
	head_tmp = listhead;
	ParseOutputPlugin(Rule);
	head_tmp=NULL;
	
#ifdef DEBUG	
	printf("***********Done Adding mandatory rules***********\n");
#endif	
}

/***********************************************************
* initialize all the snort stuff
***********************************************************/
void pass_rule_init(int argc, char* argv[]){
	//char ErrorBuff[65535];
	
	grinders[0] = DecodeEthPkt;	
	/*still needed for log_tcpdump*/
	//pd=pcap_open_live(internal_name, 1600, 0,0, ErrorBuff);

	/*from snort main*/
	InitNetmasks();
	InitProtoNames();
	
	strncpy(pv.log_dir, "/var/log/hogwash", STD_BUF-1);
	pv.log_flag=1;
 
    InitPreprocessors();
    InitPlugIns();
    InitOutputPlugins();
	
	/* setup the default rule action anchor points */
    CreateDefaultRules();
	
	strncpy(pv.log_dir, "/var/log/hogwash", STD_BUF-1);
	pv.log_flag=1;

	add_mandatory_rules();
	pass_rule_cmd_line(argc, argv);
		
	alert = OpenAlertFile("/var/log/hogwash/alert");
	
	pv.log_cmd_override=1;
	pv.data_flag=1;
	pv.show2hdr_flag=1;
	pv.logbin_flag=0;
}

/************************************************************
* The rules engine starts here
*************************************************************/
int pass_rule(PacketToProcess* ptp, int spanning){
	Packet 					p;
	int 					return_value;

#ifdef DEBUG
	printf("Got a packet\n");
#endif

#ifdef __USE_PTHREADS__
	pthread_mutex_lock(&snort_mutex);
#endif	
	FilterThis=FALSE;
    Ignore=FALSE;
#ifdef DEBUG
	printf("Calling Grinder\n");
#endif	
	//(*grinder)(&p, (struct pcap_pkthdr*)ptp, ptp->raw_packet);
	DecodeEthPkt(&p, (struct pcap_pkthdr*)ptp, ptp->raw_packet);
	p.interface_num=ptp->from_interface;

#ifdef DEBUG
	printf("Calling Preprocess\n");
#endif
	Preprocess(&p);
	ClearDumpBuf();
	return_value = !FilterThis;
#ifdef __USE_PTHREADS__	
	pthread_mutex_unlock(&snort_mutex);
#endif	
	
	if (!FilterThis){
		return FALSE;
	}	
	
	if (FilterThis && p.iph && p.tcph){
		/*send back a reset to clean up the session*/
		/* If rule that triggered is not ignore */
		if (!Ignore){
#ifdef DEBUG
			printf("This should send back a reset\n");
#endif		
			CreateResetPacket(ptp, &p, spanning);
		}
	}	
#ifdef DEBUG		
		printf("This packet should drop\n");
#endif		
	return TRUE;
}

/***************************************************************
* Be nice and free up all the stuff the malloc'd
***************************************************************/
void pass_rule_free(){
}

/***************************************************************
* Restart and re-read everything
***************************************************************/
void pass_rule_restart(){
	pv.use_rules=1;
	pv.rules_order_flag=1;
	OrderRuleLists("pass alert log");
	ParseRulesFile(pv.config_file, 0);
}
